from setuptools import setup

setup(
    name="sam_lstm",
    url="https://github.com/SheikSadi/SAM-LSTM-RESNET.git",
    author="SheikSadi",
    author_email="biis.saadi@gmail.com",
    include_package_data=True,
    packages=["sam_lstm"],
    version="1.0.1",
    description="Python library to generate saliency maps and smart cropping images",
)
